"""
cli.py
CLI for AutoLRC.
"""

import argparse
import os
import sys

import torch
import torch.cuda
import whisper

from . import lrc

VALID_WHISPER_MODELS: whisper.List[str] = whisper.available_models()


def verify_args(args_to_verify: argparse.Namespace) -> None:
    """
    Verifies arguments for the AutoLRC CLI.

    Args:
        args_to_verify (argparse.Namespace): Arguments to be verified.
    """
    if args_to_verify.model not in VALID_WHISPER_MODELS:
        raise ValueError(f"Model: '{args_to_verify.model}' is not a valid model.")
    if not os.path.exists(args_to_verify.music_dir):
        raise ValueError(f"No directory called '{args_to_verify.music_dir}' could be found.")


def initialize_parser() -> argparse.ArgumentParser:
    """
    Initializes the parser for the AutoLRC CLI.
    """
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        description="Generate LRC files for your music."
    )
    parser.add_argument(
        "music_dir",
        type=str,
        help="The directory where the music you want to generate LRC files is located.",
    )

    arg_model_help: str = f"""
    The whisper model to use for Audio To Text. Default = turbo.
    List of available modes are {VALID_WHISPER_MODELS}.
    Learn more here https://github.com/openai/whisper
    """.strip()
    parser.add_argument("--model", help=arg_model_help, default="turbo")

    return parser


def main() -> None:
    """
    Main entry point of the AutoLRC CLI.
    """

    parser: argparse.ArgumentParser = initialize_parser()
    args: argparse.Namespace = parser.parse_args()

    try:
        verify_args(args)
    except ValueError as error:
        sys.exit(str(object=error))

    model_name: str = args.model
    music_dir: str = args.music_dir

    device: str = "cuda" if torch.cuda.is_available() else "cpu"

    model: whisper.Whisper = whisper.load_model(model_name, device)

    lrc.generate_lrc_for_album(music_dir, model)


if __name__ == "__main__":
    main()



"""
lrc.py
Generates lrc files for songs using OpenAI's Whisper model.

Credits:
https://github.com/openai/whisper
https://openai.com/index/whisper/
Thanks the for inspiration:
https://www.lrcgenerator.com/
https://github.com/openai/whisper
"""

import os

import whisper
from rich.console import Console

console = Console()


def seconds_to_lrc_timestamp(seconds: float) -> str:
    """
    Converts seconds to a valid lrc timestamp.
    The format is `mm:ss:ms`.

    Args:
        seconds (float): The number of seconds.

    Examples:
        >>> seconds_to_lrc_timestamp(84)
        01:24:00
        >>> seconds_to_lrc_timestamp(583.78)
        09:43.78
        >>> seconds_to_lrc_timestamp(84.123)
        01:24.12
    """
    minutes, seconds = divmod(seconds, 60)
    milliseconds = int(round((seconds - int(seconds)) * 100))
    return f"{int(minutes):02}:{int(seconds):02}.{milliseconds:02}"


def generate_lrc(audio_path: str, model: whisper.Whisper) -> str:
    """
    Converts audio to `lrc` format.

    Args:
        audio_path (str): The path of the audio file.
        model (whisper.Whisper): The openai whisper model to be used to process the audio to text.
    """
    result = model.transcribe(audio_path)

    segments = result.get("segments")

    if not isinstance(segments, list):
        raise TypeError(
            "Expected result to have segments, try using a different model or use a different version of whisper"
        )

    lrc_result: list[str] = []

    for segment in segments:
        lrc_timestamp: str = seconds_to_lrc_timestamp(segment["start"])
        text: str = segment["text"]
        text = text.strip()

        lrc_result.append(f"[{lrc_timestamp}]{text}")

    return "\n".join(lrc_result)


def get_music_files(music_path: str) -> list[str]:
    """
    Returns a list of valid music file paths found within a directory and its subdirectories.

    Args:
        music_path (str): The file path of the directory to search.

    Returns:
        List of strings that contain file paths of valid music files.
    """
    valid_files: list[str] = []
    valid_audio_extensions: list[str] = [".mp3", ".wav"]

    for root, _, files in os.walk(music_path):
        for file in files:
            extension: str = os.path.splitext(file)[-1].lower()

            if extension in valid_audio_extensions:
                full_path: str = os.path.join(root, file)
                valid_files.append(full_path)

    return valid_files


def get_filename_without_extension(path: str) -> str:
    """
    Take a filepath and returns the name of the file without its extension.

    Args:
        path (str): Path where the file is

    Returns:
        The filename without its extension
    """
    parts: tuple[str, str] = os.path.split(path)
    last_path: str = parts[-1]
    name: str = os.path.splitext(last_path)[0]
    return name


def generate_lrc_for_album(album_path: str, model: whisper.Whisper) -> None:
    """
    Generates `lrc` files for each song in a album. (The lrc files go into the album folder)
    The `lrc` file is the same name as the songs file name with its extension replaced with `lrc`.

    Args:
        album_path (str): The file path of the album.
        model (whisper.Whisper): The openai whisper model to be used to process the audio to text.

    Raises:
        FileNotFoundError: If the path to the album doesn't exist.
    """

    if not os.path.exists(album_path):
        raise FileNotFoundError(f"No file called '{album_path}' exists.")

    status_message = "[bold green]Generating lrc files {}/{} ..."

    music_file_paths: list[str] = get_music_files(album_path)
    total_files: int = len(music_file_paths)

    console.log(f"Found {total_files} songs!")

    with console.status(status_message.format(0, total_files)) as status:
        for file_num, file in enumerate(music_file_paths, 1):
            song_name = get_filename_without_extension(file)
            lrc_path = os.path.splitext(file)[0] + "." + "lrc"

            if os.path.exists(lrc_path):
                console.log(
                    f"[bold yellow]Skipping song: [blue]{song_name}[bold yellow], cause a lrc file already exists."
                )
            else:
                console.log(f"Processing song: [blue]{song_name}")
                lrc_data: str = generate_lrc(file, model)

                with open(lrc_path, "w", encoding="utf8") as f:
                    f.write(lrc_data)

            status.update(status_message.format(file_num, total_files))



"""
cli_v2.py
CLI v2 for AutoLRC.
"""

import os
from typing import Literal

import click
import torch
import whisper

from . import lrc

VALID_WHISPER_MODELS: whisper.List[str] = whisper.available_models()

DEFAULT_MODEL: str = "turbo"

CUDA_AVAILABLE: bool = torch.cuda.is_available()
DEFAULT_TORCH_DEVICE: Literal["cuda"] | Literal["cpu"] = "cuda" if CUDA_AVAILABLE else "cpu"

if DEFAULT_MODEL not in VALID_WHISPER_MODELS:
    raise ValueError(f"Whisper model: '{DEFAULT_MODEL}' is not in {VALID_WHISPER_MODELS}")


@click.command()
@click.option(
    "--model",
    default=DEFAULT_MODEL,
    help=f"""
    The whisper model to use for Audio To Text. Default = {DEFAULT_MODEL}.
    List of available modes are {VALID_WHISPER_MODELS}.
    Learn more here https://github.com/openai/whisper""".strip(),
)
@click.option("--music-dir", prompt="Music file or Album folder")
@click.option(
    "--device",
    default="auto",
    help=f"""
    Device to run whisper on.
    Devices:
    ✔: cpu
    {"✔" if CUDA_AVAILABLE else "✖"}: cuda {"(Cuda not found)" if not CUDA_AVAILABLE else ""}
    ✔: auto (auto set to {DEFAULT_TORCH_DEVICE})
    """,
)
def cli(model: str, music_dir: str, device: str):
    if model not in VALID_WHISPER_MODELS:
        raise click.exceptions.BadParameter(f"'{model}' is not a valid model.")
    if device not in ["cpu", "cuda", "auto"]:
        raise click.exceptions.BadParameter(f"'{device}' is not a valid device.")
    if not os.path.exists(music_dir):
        raise click.exceptions.BadParameter(f"Could not find path '{music_dir}'")

    if device == "auto":
        device = DEFAULT_TORCH_DEVICE

    loaded_model: whisper.Whisper = whisper.load_model(name=model, device=device)

    lrc.generate_lrc_for_album(music_dir, loaded_model)


if __name__ == "__main__":
    cli()
